//
//                           VELOXCHEM 1.0-RC
//      ---------------------------------------------------
//                     An Electronic Structure Code
//
//  Copyright © 2018-2020 by VeloxChem developers. All rights reserved.
//  Contact: https://veloxchem.org/contact

#include<cstdint>

#include "MemBlock2D.hpp"
#include "GtoPairsBlock.hpp"
#include "RecursionMap.hpp"

namespace eribrrfunc { // eribrrfunc namespace

    /**
    Computes batch of contracted (DD|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDDXY(      CMemBlock2D<double>& braBuffer,
                                      const CRecursionMap&       recursionMap,
                                      const CMemBlock2D<double>& abDistances,
                                      const CGtoPairsBlock&      braGtoPairsBlock,
                                      const CGtoPairsBlock&      ketGtoPairsBlock,
                                      const int32_t              nKetContrPairs,
                                      const int32_t              iContrPair);

    /**
    Computes batch of contracted (DF|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDFXY(      CMemBlock2D<double>& braBuffer,
                                      const CRecursionMap&       recursionMap,
                                      const CMemBlock2D<double>& abDistances,
                                      const CGtoPairsBlock&      braGtoPairsBlock,
                                      const CGtoPairsBlock&      ketGtoPairsBlock,
                                      const int32_t              nKetContrPairs,
                                      const int32_t              iContrPair);

    /**
    Computes batch of contracted (DG|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDGXY(      CMemBlock2D<double>& braBuffer,
                                      const CRecursionMap&       recursionMap,
                                      const CMemBlock2D<double>& abDistances,
                                      const CGtoPairsBlock&      braGtoPairsBlock,
                                      const CGtoPairsBlock&      ketGtoPairsBlock,
                                      const int32_t              nKetContrPairs,
                                      const int32_t              iContrPair);

    /**
    Computes batch of contracted (DH|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDHXY(      CMemBlock2D<double>& braBuffer,
                                      const CRecursionMap&       recursionMap,
                                      const CMemBlock2D<double>& abDistances,
                                      const CGtoPairsBlock&      braGtoPairsBlock,
                                      const CGtoPairsBlock&      ketGtoPairsBlock,
                                      const int32_t              nKetContrPairs,
                                      const int32_t              iContrPair);

    /**
    Computes sub-batch (0,63) of contracted (DH|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDHXY_0_63(      CMemBlock2D<double>& braBuffer,
                                           const CRecursionMap&       recursionMap,
                                           const CMemBlock2D<double>& abDistances,
                                           const CGtoPairsBlock&      braGtoPairsBlock,
                                           const CGtoPairsBlock&      ketGtoPairsBlock,
                                           const int32_t              nKetContrPairs,
                                           const int32_t              iContrPair);

    /**
    Computes sub-batch (63,126) of contracted (DH|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDHXY_63_126(      CMemBlock2D<double>& braBuffer,
                                             const CRecursionMap&       recursionMap,
                                             const CMemBlock2D<double>& abDistances,
                                             const CGtoPairsBlock&      braGtoPairsBlock,
                                             const CGtoPairsBlock&      ketGtoPairsBlock,
                                             const int32_t              nKetContrPairs,
                                             const int32_t              iContrPair);

    /**
    Computes batch of contracted (DI|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDIXY(      CMemBlock2D<double>& braBuffer,
                                      const CRecursionMap&       recursionMap,
                                      const CMemBlock2D<double>& abDistances,
                                      const CGtoPairsBlock&      braGtoPairsBlock,
                                      const CGtoPairsBlock&      ketGtoPairsBlock,
                                      const int32_t              nKetContrPairs,
                                      const int32_t              iContrPair);

    /**
    Computes sub-batch (0,84) of contracted (DI|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDIXY_0_84(      CMemBlock2D<double>& braBuffer,
                                           const CRecursionMap&       recursionMap,
                                           const CMemBlock2D<double>& abDistances,
                                           const CGtoPairsBlock&      braGtoPairsBlock,
                                           const CGtoPairsBlock&      ketGtoPairsBlock,
                                           const int32_t              nKetContrPairs,
                                           const int32_t              iContrPair);

    /**
    Computes sub-batch (84,168) of contracted (DI|G|XY electron repulsion integrals and stores
    results in integrals buffer.

    @param braBuffer the horizontal recursion buffer for bra side.
    @param recursionMap the recursion map object.
    @param abDistances the vector of distances R(AB) = A - B.
    @param braGtoPairsBlock the GTOs pairs block on ket side.
    @param ketGtoPairsBlock the GTOs pairs block on ket side.
    @param nKetContrPairs the number of contractes GTOs pairs on ket side.
    @param iContrPair the index of contracted GTO pair on bra side.
    */
    void compElectronRepulsionForDIXY_84_168(      CMemBlock2D<double>& braBuffer,
                                             const CRecursionMap&       recursionMap,
                                             const CMemBlock2D<double>& abDistances,
                                             const CGtoPairsBlock&      braGtoPairsBlock,
                                             const CGtoPairsBlock&      ketGtoPairsBlock,
                                             const int32_t              nKetContrPairs,
                                             const int32_t              iContrPair);


} // eribrrfunc namespace

